/*
liba52 wrapper class
Copyright  2007 James Chapman
http://www.vuplayer.com

Utilizes the liba52 library
Copyright  2000-2002 Michel Lespinasse <walken@zoy.org>
Copyright  1999-2000 Aaron Holtzman <aholtzma@ess.engr.uvic.ca>
http://liba52.sourceforge.net/
*/

#include "Cac3.h"
#include <string.h>

Cac3::Cac3()
{
	m_file=0;
	m_state=a52_init(0);
}

Cac3::~Cac3()
{
	Close();
	if(m_state)
		a52_free(m_state);
}

void Cac3::Close()
{
	if(m_file)
	{
		fclose(m_file);
		m_file=0;
	}
}

bool Cac3::Open(const char* filename, float scale, bool drc)
{
	Close();
	m_scale=scale;
	m_drc=drc;
	m_file=fopen(filename,"rb");
	if(m_file)
	{
		fseek(m_file,0,SEEK_END);
		m_totalsamples=ftell(m_file);
		fseek(m_file,0,SEEK_SET);
		if(GetFrame())
		{
			m_totalsamples*=m_samplerate;
			if(m_bitrate)
				m_totalsamples/=(m_bitrate/8);
			return 1;
		}
		else
			Close();
	}
	return 0;
}

int Cac3::Read(float* buf, int bytes)
{
	int read=0;
	bytes/=4;
	while(read<bytes)
	{
		if(m_bufpos<AC3_BUFSIZE)
			buf[read++]=m_buf[m_bufpos++];
		else
		{
			if(!GetFrame())
				break;
		}
	}
	return read*4;
}

int Cac3::GetFrame()
{
	if(fread(m_frame,1,7,m_file)!=7)
		return 0;
	unsigned int framesize=a52_syncinfo(m_frame,&m_flags,&m_samplerate,&m_bitrate);
	while(!framesize)
	{
		memmove(&m_frame[0],&m_frame[1],6);
		if(!fread(&m_frame[6],1,1,m_file))
			break;
		framesize=a52_syncinfo(m_frame,&m_flags,&m_samplerate,&m_bitrate);
	}
	if(framesize>7)
	{
		if(fread(&m_frame[7],1,framesize-7,m_file)==framesize-7)
		{
			int flags=A52_STEREO;
			a52_frame(m_state,m_frame,&flags,&m_scale,0);
			if(!m_drc)
				a52_dynrng(m_state,0,0);
			for(int i=0;i<AC3_BLOCKCOUNT;i++)
			{
				a52_block(m_state);
				flags=i*AC3_BLOCKSIZE;
				for(int j=0;j<AC3_BLOCKSIZE;j++)
				{
					m_buf[(flags+j)*2]=a52_samples(m_state)[j];
					m_buf[(flags+j)*2+1]=a52_samples(m_state)[j+AC3_BLOCKSIZE];
				}
			}
		}
		else
			framesize=0;
	}
	m_bufpos=0;
	return framesize;
}
