/*
FLAC File Filter for Cool Edit/Adobe Audition
Copyright  2007 James Chapman
http://www.vuplayer.com

FLAC is Copyright  2000-2007 Josh Coalson
http://flac.sourceforge.net
*/

#include <windows.h>
#include <commctrl.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <flac/all.h>
#include "../filters.h"
#include "resource.h"

#define CHUNKSIZE	65536
#define DEFAULTMODE	5
#define PADDING		4096

#define TAGCOUNT	15

#define FCC_INFO	"INFO"
#define FCC_WAVE	"WAVE"
#define FCC_DISP	"DISP"

typedef struct
{
	LPCSTR riff;
	LPCSTR tag;
} TAGDATA;

const TAGDATA TAGS[TAGCOUNT]=
{
	{"IART","ARTIST"},
	{"ICMT","COMMENT"},
	{"ICOP","COPYRIGHT"},
	{"ICRD","DATE"},
	{"IENG","ENGINEER"},
	{"IGNR","GENRE"},
	{"IKEY","KEYWORDS"},
	{"IMED","MEDIUM"},
	{"INAM","ALBUM"},
	{"ISFT","SOFTWARE"},
	{"ISRC","SOURCE"},
	{"ITCH","TECHNICIAN"},
	{"ISBJ","SUBJECT"},
	{"ISRF","SOURCE FORM"},
	{"DISP","TITLE"}
};

typedef struct	// any special vars associated with output file
{
	WORD wChannels;
	WORD wBPS;
	BOOL skipfirst;
	char* filename;
	FLAC__int32 buffer[CHUNKSIZE];
	FLAC__StreamMetadata* tags;
	FLAC__StreamMetadata* seektable;
	FLAC__StreamMetadata* padding;
	FLAC__StreamEncoder* pEncoder;
} FILTER_OUTPUT;

typedef struct	// any special vars associated with input file
{
	DWORD dwOptions;
	DWORD bufSize;
	DWORD bufPos;
	__int64 filesize;
	FLAC__int32 buffer[FLAC__MAX_BLOCK_SIZE*4];
	FLAC__StreamMetadata flacInfo;
	FLAC__StreamMetadata* tags;
	FLAC__StreamDecoder* pDecoder;
} FILTER_INPUT;

FLAC__StreamDecoderWriteStatus writeProc(const FLAC__StreamDecoder* pDecoder,const FLAC__Frame* pFrame,const FLAC__int32 * const buffer[],void* clientData)
{
	FILTER_INPUT* pInput=(FILTER_INPUT*)clientData;
	pInput->bufSize=0;
	for(pInput->bufPos=0;pInput->bufPos<pFrame->header.blocksize;pInput->bufPos++)
	{
		for(BYTE channel=0;channel<pInput->flacInfo.data.stream_info.channels;channel++)
			pInput->buffer[pInput->bufSize++]=(buffer[channel][pInput->bufPos]);
	}
	pInput->bufPos=0;
	return FLAC__STREAM_DECODER_WRITE_STATUS_CONTINUE;
}

void metadataProc(const FLAC__StreamDecoder* pDecoder,const FLAC__StreamMetadata* metadata, void* clientData)
{
}

void errorProc(const FLAC__StreamDecoder* pDecoder, FLAC__StreamDecoderErrorStatus status, void* clientData)
{
}

BOOL GetNextComment(FLAC__StreamMetadata* metadata, SPECIALDATA* psp);
BOOL SetComment(FLAC__StreamMetadata* metadata, char* field, char* value);
BOOL ANSItoUTF8(char* ansi, char** utf8);
BOOL UTF8toANSI(char* utf8, char** ansi);
void EnsureCRLF(WCHAR** str);

//////////////////////////////////////////////////////////////////////
// REQUIRED COOL EDIT FILTER FUNCTIONS
//////////////////////////////////////////////////////////////////////

__declspec(dllexport) short FAR PASCAL QueryCoolFilter(COOLQUERY* cq)
{
	ZeroMemory(cq,sizeof(COOLQUERY));
	strcpy(cq->szName,"FLAC");		
	strcpy(cq->szCopyright,"Free Lossless Audio Codec 1.2.1");
	strcpy(cq->szExt,"FLA");
	strcpy(cq->szExt2,"FL*");
	cq->lChunkSize=CHUNKSIZE; 
	cq->dwFlags=QF_CANLOAD|QF_CANSAVE|QF_RATEADJUSTABLE|QF_HASOPTIONSBOX|QF_CANDO32BITFLOATS|QF_READSPECIALFIRST|QF_WRITESPECIALFIRST;
 	cq->Stereo8=0xFF;
 	cq->Stereo16=0xFF;
 	cq->Stereo32=0xFF;
 	cq->Mono8=0xFF;
 	cq->Mono16=0xFF;
 	cq->Mono32=0xFF;
 	return C_VALIDLIBRARY;
}

__declspec(dllexport) BOOL FAR PASCAL FilterUnderstandsFormat(LPSTR filename)
{
	FLAC__StreamMetadata flacInfo;
	if(FLAC__metadata_get_streaminfo(filename,&flacInfo))
		return 1;
	else
		return 0;	
}

//////////////////////////////////////////////////////////////////////
// REQUIRED COOL EDIT FILTER FUNCTIONS - READING
//////////////////////////////////////////////////////////////////////

__declspec(dllexport) HANDLE FAR PASCAL OpenFilterInput(LPSTR filename,long* lRate,WORD* wBPS,WORD* wChannels,HWND hWnd,long* lChunkSize)
{
	FILTER_INPUT *filter=new FILTER_INPUT;
	if(filter)
	{
		ZeroMemory(filter,sizeof(FILTER_INPUT));
		struct _stati64 s;
		_stati64(filename,&s);
		filter->filesize=s.st_size;
		FLAC__metadata_get_streaminfo(filename,&filter->flacInfo);
		switch(filter->flacInfo.data.stream_info.bits_per_sample)
		{
		case 24:
			*wBPS=32;
			break;
		case 16:
			*wBPS=16;
			break;
		case 8:
			*wBPS=8;
			break;
		default:
			delete filter;
			return 0;
		}
		*lRate=filter->flacInfo.data.stream_info.sample_rate;
		*wChannels=filter->flacInfo.data.stream_info.channels;
		FLAC__metadata_get_tags(filename,&filter->tags);
		filter->pDecoder=FLAC__stream_decoder_new();
		if(filter->pDecoder)
		{
			FLAC__stream_decoder_set_metadata_ignore_all(filter->pDecoder);
			FLAC__stream_decoder_set_metadata_respond(filter->pDecoder,FLAC__METADATA_TYPE_STREAMINFO);
			if(FLAC__stream_decoder_init_file(filter->pDecoder,filename,writeProc,metadataProc,errorProc,filter)==FLAC__STREAM_DECODER_INIT_STATUS_OK)
			{
				FLAC__stream_decoder_process_until_end_of_metadata(filter->pDecoder);
				return filter;
			}
			FLAC__stream_decoder_delete(filter->pDecoder);
		}
		delete filter;
	}
	return 0;
}

__declspec(dllexport) DWORD FAR PASCAL FilterGetFileSize(HANDLE hInput)
{
	FILTER_INPUT *filter=(FILTER_INPUT*)hInput;
	__int64 size=filter->flacInfo.data.stream_info.total_samples
			*((filter->flacInfo.data.stream_info.bits_per_sample==24)
			?4:filter->flacInfo.data.stream_info.bits_per_sample/8)
			*filter->flacInfo.data.stream_info.channels;
	if(size>0xffffffff)
	{
		size=0x100000000;
		size-=((filter->flacInfo.data.stream_info.bits_per_sample==24)
			?4:filter->flacInfo.data.stream_info.bits_per_sample/8)
			*filter->flacInfo.data.stream_info.channels;
	}
	return (DWORD)size;
}

__declspec(dllexport) DWORD FAR PASCAL ReadFilterInput(HANDLE hInput, unsigned char* buf, long lBytes)
{
	FILTER_INPUT *filter=(FILTER_INPUT*)hInput;
	long lRead=0;
	float* pFloat=(float*)buf;
	short* pShort=(short*)buf;
	while(lRead<lBytes)
	{
		if(filter->bufPos<filter->bufSize)
		{
			switch(filter->flacInfo.data.stream_info.bits_per_sample)
			{
			case 24:
				pFloat[lRead/4]=(float)filter->buffer[filter->bufPos];
				pFloat[lRead/4]/=256;
				lRead+=4;
				break;
			case 16:
				pShort[lRead/2]=(short)filter->buffer[filter->bufPos];
				lRead+=2;
				break;
			case 8:
				buf[lRead]=(unsigned char)(filter->buffer[filter->bufPos]+128);
				lRead++;
			}
			filter->bufPos++;
		}
		else
		{
			if((FLAC__stream_decoder_process_single(filter->pDecoder)==false)||!filter->bufSize)
				break;
		}
	}
	return lRead;
}

__declspec(dllexport) void FAR PASCAL CloseFilterInput(HANDLE hInput)
{
	FILTER_INPUT *filter=(FILTER_INPUT*)hInput;
	if(filter)
	{
		if(filter->pDecoder)
		{
			FLAC__stream_decoder_finish(filter->pDecoder);
			FLAC__stream_decoder_delete(filter->pDecoder);
			if(filter->tags)
				FLAC__metadata_object_delete(filter->tags);
		}
		delete filter;
	}
}

//////////////////////////////////////////////////////////////////////
// REQUIRED COOL EDIT FILTER FUNCTIONS - WRITING
//////////////////////////////////////////////////////////////////////

__declspec(dllexport) void FAR PASCAL GetSuggestedSampleType(long* lRate, WORD* wBPS, WORD *wChannels)
{
	*wBPS=0;
	*lRate=0;
	*wChannels=0;
}

__declspec(dllexport) HANDLE FAR PASCAL OpenFilterOutput(LPSTR filename,
														 long samprate,
														 WORD BPS,
														 WORD channels,
														 DWORD lSize,
														 long* lChunkSize,
														 DWORD options)
{
	FILTER_OUTPUT *filter=new FILTER_OUTPUT;
	if(!filter)
		return 0;
	ZeroMemory(filter,sizeof(FILTER_OUTPUT));
	filter->pEncoder=FLAC__stream_encoder_new();
	if(!filter->pEncoder)
	{
		delete filter;
		return 0;
	}
	*lChunkSize=CHUNKSIZE;
	filter->wChannels=channels;
	filter->wBPS=BPS;
	if(BPS==32)
		BPS=24;

	// Force .flac file extension
	filter->filename=new char[strlen(filename)+6];
	ZeroMemory(filter->filename,strlen(filename)+6);
	char* a=strrchr(filename,'.');
	if(a)
	{
		strncpy(filter->filename,filename,a-filename);
		a=strrchr(filter->filename,'.');
		if(a&&!_stricmp(a,".flac"))
			*a=0;
	}
	else
		strcpy(filter->filename,filename);
	strcat(filter->filename,".flac");

	FLAC__stream_encoder_set_channels(filter->pEncoder,channels);
	FLAC__stream_encoder_set_bits_per_sample(filter->pEncoder,BPS);
	FLAC__stream_encoder_set_sample_rate(filter->pEncoder,samprate);
	FLAC__stream_encoder_set_total_samples_estimate(filter->pEncoder,lSize/channels/(BPS/8));
	FLAC__stream_encoder_set_compression_level(filter->pEncoder,(options-1<=8)?options-1:DEFAULTMODE);
	filter->tags=FLAC__metadata_object_new(FLAC__METADATA_TYPE_VORBIS_COMMENT);
	filter->padding=FLAC__metadata_object_new(FLAC__METADATA_TYPE_PADDING);
	filter->seektable=FLAC__metadata_object_new(FLAC__METADATA_TYPE_SEEKTABLE);
	if(filter->padding)
		filter->padding->length=PADDING;
	if(filter->seektable)
		FLAC__metadata_object_seektable_template_append_spaced_points(filter->seektable,100,lSize/channels/(BPS/8));
	return filter;
}

__declspec(dllexport) DWORD FAR PASCAL WriteFilterOutput(HANDLE hOutput, BYTE* pBuf, long lBytes)
{
	FILTER_OUTPUT *filter=(FILTER_OUTPUT*)hOutput;
	if(!filter->skipfirst)
	{
		filter->skipfirst=1;
		FLAC__StreamMetadata* metadata[3];
		metadata[0]=filter->tags;
		metadata[1]=filter->padding;
		metadata[2]=filter->seektable;
		FLAC__stream_encoder_set_metadata(filter->pEncoder,metadata,3);
		if(FLAC__stream_encoder_init_file(filter->pEncoder,filter->filename,0,filter)!=FLAC__STREAM_ENCODER_INIT_STATUS_OK)
			return 0;		
	}
	switch(filter->wBPS)
	{
	case 32:
		{
			float* pFloat=(float*)pBuf;
			for(int index=0;index<lBytes/4;index++)
			{
				filter->buffer[index]=(FLAC__int32)(pFloat[index]*256);
				if(filter->buffer[index]>8388607)
					filter->buffer[index]=8388607;
				else
				{
					if(filter->buffer[index]<-8388608)
						filter->buffer[index]=-8388608;
				}
			}
		}
		break;
	case 16:
		{
			short* pShort=(short*)pBuf;
			for(int index=0;index<lBytes/2;index++)
				filter->buffer[index]=pShort[index];
		}
		break;
	case 8:
		{
			for(int index=0;index<lBytes;index++)
				filter->buffer[index]=pBuf[index]+128;
		}
	}
	if(FLAC__stream_encoder_process_interleaved(filter->pEncoder,filter->buffer,lBytes/filter->wChannels/(filter->wBPS/8)))
		return lBytes;
	else
		return 0;
}

__declspec(dllexport) void FAR PASCAL CloseFilterOutput(HANDLE hOutput)
{
	FILTER_OUTPUT *filter=(FILTER_OUTPUT*)hOutput;
	if(filter)
	{
		if(filter->pEncoder)
		{
			FLAC__stream_encoder_finish(filter->pEncoder);
			FLAC__stream_encoder_delete(filter->pEncoder);
			if(filter->tags)
				FLAC__metadata_object_delete(filter->tags);
			if(filter->padding)
				FLAC__metadata_object_delete(filter->padding);
			if(filter->seektable)
				FLAC__metadata_object_delete(filter->seektable);
		}
		delete [] filter->filename;
		delete filter;
	}
}

//////////////////////////////////////////////////////////////////////
// OPTIONAL COOL EDIT FILTER FUNCTIONS
//////////////////////////////////////////////////////////////////////

__declspec(dllexport) DWORD FAR PASCAL FilterOptionsString(HANDLE hInput, LPSTR lpszString)
{
	FILTER_INPUT *filter=(FILTER_INPUT*)hInput;
	if(filter)
	{
		char str[128]="FLAC";
		char* a=(filter->tags)?(strstr((char*)filter->tags->data.vorbis_comment.vendor_string.entry,str)):0;
		if(a)
			strncat(str,a+4,96);
		__int64 bytes=filter->flacInfo.data.stream_info.total_samples
			*filter->flacInfo.data.stream_info.channels
			*(filter->flacInfo.data.stream_info.bits_per_sample/8);
		if(bytes)
		{
			float f=(float)filter->filesize;
			f/=bytes;
			char buffer[32]="";
			sprintf(buffer,"\nCompression ratio: %.3f",f);
			strcat(str,buffer);
		}
		strcpy(lpszString,str);
	}
	return 0;
}

__declspec(dllexport) DWORD FAR PASCAL FilterGetOptions(HWND hWnd,
														HINSTANCE hInst,
														long lSamprate,
														WORD wChannels,
														WORD wBPS,
														DWORD dwOptions)
{
	FARPROC lpfnDialogMsgProc=GetProcAddress(hInst,(LPCSTR)"DIALOGMsgProc");
	DWORD nDialogReturn=(DWORD)DialogBoxParam((HINSTANCE)hInst,(LPCSTR)MAKEINTRESOURCE(IDD_CONFIG),hWnd,(DLGPROC)lpfnDialogMsgProc,dwOptions);
	if(nDialogReturn>0)
		return nDialogReturn;
	else
		return dwOptions;
}

__declspec(dllexport) DWORD FAR PASCAL FilterOptions(HANDLE hInput)
{
	FILTER_INPUT *filter=(FILTER_INPUT*)hInput;
	if(!filter)
		return 0;
	return filter->dwOptions;
}

__declspec(dllexport) DWORD FAR PASCAL FilterWriteSpecialData(HANDLE hOutput, LPCSTR szListType, LPCSTR szType, char * pData, DWORD dwSize)
{
	FILTER_OUTPUT *filter=(FILTER_OUTPUT*)hOutput;
	if(filter&&filter->tags)
	{
		if((!strcmp(szListType,FCC_INFO))||(!strcmp(szListType,FCC_WAVE)&&!strcmp(szType,FCC_DISP)&&dwSize>4))
		{
			char* utf;
			if(ANSItoUTF8(strcmp(szListType,FCC_INFO)?pData+4:pData,&utf))
			{
				for(int i=0;i<TAGCOUNT;i++)
				{
					if(!strcmp(szType,TAGS[i].riff))
					{
						SetComment(filter->tags,(char*)TAGS[i].tag,utf);
						break;
					}
				}
				delete [] utf;
			}
		}
	}
	return 1;
}

__declspec(dllexport) DWORD FAR PASCAL FilterGetFirstSpecialData(HANDLE hInput,	SPECIALDATA * psp)
{
	FILTER_INPUT *filter=(FILTER_INPUT*)hInput;
	if(filter&&filter->tags)
	{
		psp->hSpecialData=0;
		if(GetNextComment(filter->tags,psp))
			return 1;
	}
	return 0;
}

__declspec(dllexport) DWORD FAR PASCAL FilterGetNextSpecialData(HANDLE hInput, SPECIALDATA * psp)
{
	FILTER_INPUT *filter=(FILTER_INPUT*)hInput;
	if(filter&&filter->tags)
	{
		if(GetNextComment(filter->tags,psp))
			return 1;
	}
	return 0;
}

__declspec(dllexport) BOOL FAR PASCAL DIALOGMsgProc(HWND hWndDlg, UINT Message, WPARAM wParam, LPARAM lParam)
{
	switch(Message)
	{
	case WM_INITDIALOG:
		{
			SendDlgItemMessage(hWndDlg,IDC_QUALITYSLIDER,TBM_SETRANGEMIN,0,0);
			SendDlgItemMessage(hWndDlg,IDC_QUALITYSLIDER,TBM_SETRANGEMAX,0,8);
			SendDlgItemMessage(hWndDlg,IDC_QUALITYSLIDER,TBM_SETTICFREQ,1,0);
			SendDlgItemMessage(hWndDlg,IDC_QUALITYSLIDER,TBM_SETPAGESIZE,0,1);
			WORD mode=lParam-1;
			if(mode>8)
				SendDlgItemMessage(hWndDlg,IDC_QUALITYSLIDER,TBM_SETPOS,1,DEFAULTMODE);
			else
				SendDlgItemMessage(hWndDlg,IDC_QUALITYSLIDER,TBM_SETPOS,1,mode);
		}
		return 1;
	case WM_CLOSE:
		PostMessage(hWndDlg,WM_COMMAND,IDCANCEL,0);
		return 1;
	case WM_NOTIFY:
		{
			LPNMHDR pnmh=(LPNMHDR)lParam;
			if(pnmh->idFrom==IDC_QUALITYSLIDER)
			{
				char buffer[2]="";
				int pos=SendDlgItemMessage(hWndDlg,IDC_QUALITYSLIDER,TBM_GETPOS,0,0);
				itoa(pos,buffer,10);
				SetDlgItemText(hWndDlg,IDC_QUALITY,buffer);
				return 1;
			}
		}
		break;
	case WM_COMMAND:
		switch(LOWORD(wParam))
		{
		case IDOK:
			EndDialog(hWndDlg,SendDlgItemMessage(hWndDlg,IDC_QUALITYSLIDER,TBM_GETPOS,0,0)+1);
			return 1;
		case IDCANCEL:
			EndDialog(hWndDlg,0);
			return 1;
		}
 	}
	return 0;
}

//////////////////////////////////////////////////////////////////////
// INTERNAL FUNCTIONS
//////////////////////////////////////////////////////////////////////

BOOL GetNextComment(FLAC__StreamMetadata* metadata, SPECIALDATA* psp)
{
	for(UINT i=(UINT)(void*)psp->hSpecialData;i<metadata->data.vorbis_comment.num_comments;i++)
	{
		psp->hSpecialData=(void*)(i+1);
		char* e=strchr((char*)(metadata->data.vorbis_comment.comments[i].entry),'=');
		if(e&&e-(char*)(metadata->data.vorbis_comment.comments[i].entry)&&strlen(e+1))
		{
			char* t=new char[e-(char*)(metadata->data.vorbis_comment.comments[i].entry)+1];
			if(t)
			{
				ZeroMemory(t,e-(char*)(metadata->data.vorbis_comment.comments[i].entry)+1);
				strncpy(t,(char*)(metadata->data.vorbis_comment.comments[i].entry),e-(char*)(metadata->data.vorbis_comment.comments[i].entry));
				for(int j=0;j<TAGCOUNT;j++)
				{
					if(!_stricmp(t,TAGS[j].tag))
					{
						char* ansi;
						if(UTF8toANSI(e+1,&ansi))
						{
							int x=strlen(ansi)+2;
							BOOL bDisp=0;
							if(!strcmp(TAGS[j].riff,FCC_DISP))
							{
								bDisp=1;
								x+=4;
							}
							psp->hData=GlobalAlloc(GMEM_MOVEABLE|GMEM_ZEROINIT,x);
							if(psp->hData)
							{
								char* pData=(char*)GlobalLock(psp->hData);
								psp->dwSize=x-1;
								psp->dwExtra=1;
								if(bDisp)
								{
									strcpy(psp->szListType,FCC_WAVE);
									strcpy(pData+4,ansi);
									pData[0]=1;
								}
								else
								{
									strcpy(psp->szListType,FCC_INFO);
									strcpy(pData,ansi);
								}
								strcpy(psp->szType,TAGS[j].riff);
								GlobalUnlock(pData);
								delete [] ansi;
								delete [] t;
								return 1;
							}
							delete [] ansi;
						}
						break;
					}
				}
				delete [] t;
			}
		}
	}
	return 0;
}

BOOL SetComment(FLAC__StreamMetadata* metadata, char* field, char* value)
{
	BOOL success=0;
	char* comment=new char[strlen(field)+strlen(value)+2];
	if(comment)
	{
		strcpy(comment,field);
		strcat(comment,"=");
		strcat(comment,value);
		FLAC__StreamMetadata_VorbisComment_Entry vc;
		vc.length=strlen(comment);
		vc.entry=(FLAC__byte*)comment;
		success=FLAC__metadata_object_vorbiscomment_append_comment(metadata,vc,1);
		delete [] comment;
	}
	return success;
}

BOOL ANSItoUTF8(char* ansi, char** utf8)
{
	BOOL success=0;
	int i=MultiByteToWideChar(CP_ACP,MB_PRECOMPOSED,ansi,-1,0,0);
	if(i>1)
	{
		WCHAR* wBuf=new WCHAR[i];
		if(wBuf)
		{
			MultiByteToWideChar(CP_ACP,MB_PRECOMPOSED,ansi,-1,wBuf,i);
			i=WideCharToMultiByte(CP_UTF8,0,wBuf,-1,0,0,0,0);
			if(i>1)
			{
				char* buf=new char[i];
				if(buf)
				{
					WideCharToMultiByte(CP_UTF8,0,wBuf,-1,buf,i,0,0);
					*utf8=buf;
					success=1;
				}
			}
			delete [] wBuf;
		}
	}
	return success;
}

BOOL UTF8toANSI(char* utf8, char** ansi)
{
	BOOL success=0;
	int i=MultiByteToWideChar(CP_UTF8,0,utf8,-1,0,0);
	if(i>1)
	{
		WCHAR* wBuf=new WCHAR[i];
		if(wBuf)
		{
			MultiByteToWideChar(CP_UTF8,0,utf8,-1,wBuf,i);
			EnsureCRLF(&wBuf);
			i=WideCharToMultiByte(CP_ACP,0,wBuf,-1,0,0,0,0);
			if(i>1)
			{
				char* buf=new char[i];
				if(buf)
				{
					WideCharToMultiByte(CP_ACP,0,wBuf,-1,buf,i,0,0);
					*ansi=buf;
					success=1;
				}
			}
			delete [] wBuf;
		}
	}
	return success;
}

void EnsureCRLF(WCHAR** str)
{
	WCHAR* a=*str;
	WCHAR* b;
	int i;
	while(b=wcsstr(a,L"\r\n"))
	{
		a=++b+1;
		i=wcslen(a);
		if(i)
			memmove(b,a,i*2);
		*(b+i)=0;
	}
	a=*str;
	i=0;
	while(b=wcspbrk(a,L"\r\n"))
	{
		a=b+1;
		i++;
	}
	if(i)
	{
		a=*str;
		i=wcslen(a)+i+1;
		WCHAR* w=new WCHAR[i];
		if(w)
		{
			ZeroMemory(w,i*sizeof(WCHAR));
			while(b=wcspbrk(a,L"\r\n"))
			{
				wcsncat(w,a,b-a);
				wcscat(w,L"\r\n");
				a=b+1;
			}
			wcscat(w,a);
			delete [] *str;
			*str=w;
		}
	}
}
