/*
MOD Music File Filter for Cool Edit/Adobe Audition
Copyright  2006 James Chapman
http://www.vuplayer.com

BASS library is Copyright  1999-2006 Ian Luck
http://www.un4seen.com
*/

#include <windows.h>
#include <commctrl.h>
#include <stdio.h>
#include "../filters.h"
#include "resource.h"
#include "bass.h"

#define CHUNKSIZE			65536
#define NONINTERPOLATE		0x800000

const int DEFAULT_PANNING=50;
const int DEFAULT_RAMPING=2;
const int DEFAULT_SURROUND=0;
const int DEFAULT_LOAD=0;

char* OPTIONS_RAMPING[3]={"Off","Normal","Sensitive"};
char* OPTIONS_SURROUND[3]={"Off","Mode 1","Mode 2"};
char* OPTIONS_LOAD[3]={"Normal","FastTracker 2","ProTracker 1"};

typedef struct	// any special vars associated with input file
{
	DWORD options;
	BOOL init;
	HMUSIC hmusic;
} FILTER_INPUT;

//////////////////////////////////////////////////////////////////////
// REQUIRED COOL EDIT FILTER FUNCTIONS
//////////////////////////////////////////////////////////////////////

__declspec(dllexport) short FAR PASCAL QueryCoolFilter(COOLQUERY* cq)
{
	ZeroMemory(cq,sizeof(COOLQUERY));
	strcpy(cq->szName,"MOD Music");		
	strcpy(cq->szCopyright,"BASS library  un4seen developments");
	strcpy(cq->szExt,"mod");
	strcpy(cq->szExt2,"s3m");
	strcpy(cq->szExt3,"xm");
	strcpy(cq->szExt4,"it");
	cq->lChunkSize=CHUNKSIZE;
	cq->dwFlags=QF_CANLOAD|QF_CANDO32BITFLOATS|QF_HASOPTIONSBOX;
	cq->Stereo32=0xFF;
 	return C_VALIDLIBRARY;
}

__declspec(dllexport) BOOL FAR PASCAL FilterUnderstandsFormat(LPSTR filename)
{
	BOOL success=0;
	if(BASS_Init(0,44100,0,0,0))
	{
		if(BASS_MusicLoad(0,filename,0,0,BASS_SAMPLE_FLOAT|BASS_MUSIC_DECODE|BASS_MUSIC_NOSAMPLE,0))
			success=1;
		BASS_Free();
	}
	return success;	
}

//////////////////////////////////////////////////////////////////////
// REQUIRED COOL EDIT FILTER FUNCTIONS - READING
//////////////////////////////////////////////////////////////////////

__declspec(dllexport) HANDLE FAR PASCAL OpenFilterInput(LPSTR filename,long* lRate,WORD* wBPS,WORD* wChannels,HWND hWnd,long* lChunkSize)
{
	if(BASS_Init(0,44100,0,0,0))
	{
		FILTER_INPUT* filter=new FILTER_INPUT;
		if(filter)
		{
			filter->init=1;
			filter->hmusic=BASS_MusicLoad(0,filename,0,0,BASS_SAMPLE_FLOAT|BASS_MUSIC_DECODE|BASS_MUSIC_PRESCAN,0);
			if(filter->hmusic)
			{
				*lRate=44100;
				*wBPS=32;
				*wChannels=2;
				*lChunkSize=CHUNKSIZE;
				return filter;
			}
			delete filter;
		}
		BASS_Free();
	}
	return 0;
}

__declspec(dllexport) DWORD FAR PASCAL FilterGetFileSize(HANDLE hInput)
{
	FILTER_INPUT *filter=(FILTER_INPUT*)hInput;
	return (DWORD)BASS_ChannelGetLength(filter->hmusic);
}

__declspec(dllexport) DWORD FAR PASCAL ReadFilterInput(HANDLE hInput, unsigned char* buf, long lBytes)
{
	FILTER_INPUT *filter=(FILTER_INPUT*)hInput;
	if(filter->init)
	{
		filter->init=0;
		BASS_CHANNELINFO info;
		BASS_ChannelGetInfo(filter->hmusic,&info);
		DWORD flags=info.flags;
		filter->options-=1;
		if((filter->options&0xff)<=100)
			BASS_MusicSetAttribute(filter->hmusic,BASS_MUSIC_ATTRIB_PANSEP,filter->options&0xff);
		switch(filter->options>>8&0x03)
		{
		case 1:
			flags|=BASS_MUSIC_RAMP;
			break;
		case 2:
			flags|=BASS_MUSIC_RAMPS;
		}
		switch(filter->options>>10&0x03)
		{
		case 1:
			flags|=BASS_MUSIC_SURROUND;
			break;
		case 2:
			flags|=BASS_MUSIC_SURROUND2;
		}
		switch(filter->options>>12&0x03)
		{
		case 1:
			flags|=BASS_MUSIC_FT2MOD;
			break;
		case 2:
			flags|=BASS_MUSIC_PT1MOD;
		}
		if(filter->options&NONINTERPOLATE)
			flags|=BASS_MUSIC_NONINTER;
		BASS_ChannelSetFlags(filter->hmusic,flags);
	}
	float* f=(float*)buf;
	lBytes=BASS_ChannelGetData(filter->hmusic,buf,lBytes);
	for(long i=0;i<lBytes/4;i++)
		f[i]*=32768.0f;
	return lBytes;
}

__declspec(dllexport) void FAR PASCAL CloseFilterInput(HANDLE hInput)
{
	FILTER_INPUT *filter=(FILTER_INPUT*)hInput;
	if(filter)
	{
		BASS_Free();
		delete filter;
	}
}

//////////////////////////////////////////////////////////////////////
// OPTIONAL COOL EDIT FILTER FUNCTIONS
//////////////////////////////////////////////////////////////////////

__declspec(dllexport) DWORD FAR PASCAL FilterGetOptions(HWND hWnd,
														HINSTANCE hInst,
														long lSamprate,
														WORD wChannels,
														WORD wBPS,
														DWORD dwOptions)
{
	FARPROC lpfnDIALOGMsgProc=GetProcAddress(hInst,(LPCSTR)"DIALOGMsgProc");
	long nDialogReturn=DialogBoxParam((HINSTANCE)hInst,(LPCSTR)MAKEINTRESOURCE(IDD_OPTIONS),hWnd,(DLGPROC)lpfnDIALOGMsgProc,dwOptions);
	if(nDialogReturn>0)
		return nDialogReturn;
	else
		return 0;
}

__declspec(dllexport) DWORD FAR PASCAL FilterSetOptions(HANDLE hInput, DWORD dwOptions, long lSamprate, WORD wChannels, WORD wBPS)
{
	FILTER_INPUT *filter=(FILTER_INPUT*)hInput;
	if(filter)
		filter->options=dwOptions;
	return 0;
}

__declspec(dllexport) DWORD FAR PASCAL FilterOptions(HANDLE hInput)
{
	return 0;
}

__declspec(dllexport) DWORD FAR PASCAL FilterOptionsString(HANDLE hInput, LPSTR lpszString)
{
	FILTER_INPUT *filter=(FILTER_INPUT*)hInput;
	if(filter&&filter->hmusic)
	{
		BASS_CHANNELINFO info;
		BASS_ChannelGetInfo(filter->hmusic,&info);
		if(info.ctype&BASS_CTYPE_MUSIC_MO3)
			info.ctype^=BASS_CTYPE_MUSIC_MO3;
		switch(info.ctype)
		{
		case BASS_CTYPE_MUSIC_MOD:
			strcpy(lpszString,"ProTracker");
			break;
		case BASS_CTYPE_MUSIC_MTM:
			strcpy(lpszString,"MultiTracker");
			break;
		case BASS_CTYPE_MUSIC_S3M:
			strcpy(lpszString,"ScreamTracker");
			break;
		case BASS_CTYPE_MUSIC_XM:
			strcpy(lpszString,"FastTracker");
			break;
		case BASS_CTYPE_MUSIC_IT:
			strcpy(lpszString,"ImpulseTracker");
		}
		int i=0;
		while(BASS_MusicGetAttribute(filter->hmusic,BASS_MUSIC_ATTRIB_VOL_CHAN+i)!=-1)
			i++;  
		char buffer[16]="";
		if(i)
		{
			sprintf(buffer,"\nChannels: %d",i);
			strcat(lpszString,buffer);
		}
		i=0;
		while(BASS_MusicGetAttribute(filter->hmusic,BASS_MUSIC_ATTRIB_VOL_INST+i)!=-1)
			i++;
		if(i)
		{
			sprintf(buffer,"\nInstruments: %d",i);
			strcat(lpszString,buffer);
		}
	}
	return 0;
}

__declspec(dllexport) BOOL FAR PASCAL DIALOGMsgProc(HWND hWndDlg, UINT Message, WPARAM wParam, LPARAM lParam)
{
	switch(Message)
	{
	case WM_INITDIALOG:
		{
			DWORD i;
			SendDlgItemMessage(hWndDlg,IDC_PANSLIDER,TBM_SETRANGEMIN,0,0);
			SendDlgItemMessage(hWndDlg,IDC_PANSLIDER,TBM_SETRANGEMAX,0,100);
			SendDlgItemMessage(hWndDlg,IDC_PANSLIDER,TBM_SETTICFREQ,10,0);
			SendDlgItemMessage(hWndDlg,IDC_PANSLIDER,TBM_SETPAGESIZE,0,10);
			for(i=0;i<3;i++)
				SendDlgItemMessage(hWndDlg,IDC_RAMPING,CB_ADDSTRING,0,(LPARAM)OPTIONS_RAMPING[i]);
			for(i=0;i<3;i++)
				SendDlgItemMessage(hWndDlg,IDC_SURROUND,CB_ADDSTRING,0,(LPARAM)OPTIONS_SURROUND[i]);
			for(i=0;i<3;i++)
				SendDlgItemMessage(hWndDlg,IDC_LOAD,CB_ADDSTRING,0,(LPARAM)OPTIONS_LOAD[i]);
			SendDlgItemMessage(hWndDlg,IDC_DEFAULT,BM_SETCHECK,BST_CHECKED,0);
			SendDlgItemMessage(hWndDlg,IDC_PANSLIDER,TBM_SETPOS,1,DEFAULT_PANNING);
			SendDlgItemMessage(hWndDlg,IDC_RAMPING,CB_SETCURSEL,DEFAULT_RAMPING,0);
			SendDlgItemMessage(hWndDlg,IDC_SURROUND,CB_SETCURSEL,DEFAULT_SURROUND,0);
			SendDlgItemMessage(hWndDlg,IDC_LOAD,CB_SETCURSEL,DEFAULT_LOAD,0);
			RECT rect;
			GetClientRect(hWndDlg,&rect);
			SetWindowPos(hWndDlg,0,(GetSystemMetrics(SM_CXSCREEN)-rect.right)/2,(GetSystemMetrics(SM_CYSCREEN)-rect.bottom)/2,0,0,SWP_NOSIZE|SWP_NOZORDER);
		}
		return TRUE;
    case WM_CLOSE:
		PostMessage(hWndDlg,WM_COMMAND,IDCANCEL,0L);
		return TRUE;
	case WM_NOTIFY:
		{
			LPNMHDR pnmh=(LPNMHDR)lParam;
			if(pnmh->idFrom==IDC_PANSLIDER)
			{
				char buffer[16]="";
				itoa(SendDlgItemMessage(hWndDlg,IDC_PANSLIDER,TBM_GETPOS,0,0),buffer,10);
				SetDlgItemText(hWndDlg,IDC_PANVALUE,buffer);
			}
		}
		break;
    case WM_COMMAND:
		{
			DWORD options=0;
			switch(LOWORD(wParam))
			{
			case IDOK:
				options=SendDlgItemMessage(hWndDlg,IDC_LOAD,CB_GETCURSEL,0,0);
				options<<=2;
				options+=SendDlgItemMessage(hWndDlg,IDC_SURROUND,CB_GETCURSEL,0,0);
				options<<=2;
				options+=SendDlgItemMessage(hWndDlg,IDC_RAMPING,CB_GETCURSEL,0,0);
				options<<=8;
				options+=SendDlgItemMessage(hWndDlg,IDC_PANSLIDER,TBM_GETPOS,0,0);
				if(SendDlgItemMessage(hWndDlg,IDC_INTERPOLATION,BM_GETCHECK,0,0)==BST_CHECKED)
					options|=NONINTERPOLATE;
				options+=1;
			case IDCANCEL:
				EndDialog(hWndDlg,options);
				return TRUE;
			}
		}
	}
	return FALSE;
}
